(* An interpreter for arithmetic expressions with integers and addition,
   and let bindings. *)

(* The type of the abstract syntax tree (AST),
 * that is, expresions of the language. *)
type expr =
  | Var of string
  | Int of int
  | Add of expr * expr
  | Let of string * expr * expr

(* [is_value e] is whether [e] is a syntactic value *)
let is_value : expr -> bool = function
  | Int _ -> true
  | Add _ | Var _ | Let _ -> false

(* [subst e v x] is e{v/x}, that is, [e] with [v]
 * substituted for [x]. *)
let rec subst e v x = match e with
  | Var y      -> if x=y then v else e
  | Int n      -> Int n
  | Add(el,er) -> Add(subst el v x, subst er v x)
  | Let(y,ebind,ebody) ->
      let ebind' = subst ebind v x in
      if x=y
      then Let(y, ebind', ebody)
      else Let(y, ebind', subst ebody v x)

(* A single step of evaluation: exactly 1 step *)
let rec step : expr -> expr = function
  | Int n               -> failwith "Does not step"
  | Var _               -> failwith "Unbound variable"
  | Add(Int n1, Int n2) -> Int (n1 + n2)
  | Add(Int n1, e2)     -> Add (Int n1, step e2)
  | Add(e1, e2)         -> Add (step e1, e2)
  | Let(x, Int n, e2)   -> subst e2 (Int n) x
  | Let(x, e1, e2)      -> Let (x, step e1, e2)

(* [eval e] is the [e -->* v] judgement.  That is,
 * keep applying [step] until a value is produced.  *)
let rec eval : expr -> expr = fun e ->
  if is_value e then e
  else eval (step e)

